/*-
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package log

/*
#cgo pkg-config: varnishapi
#include <stdio.h>
#include <vapi/vsl.h>
*/
import "C"

import (
	"bytes"
	"errors"
)

var spaceByte = []byte(" ")

const (
	cutoff = ^uint32(0) / 10
	rest   = byte(^uint32(0)-cutoff*10) + byte('0')
)

// Calling strconv.Atoi(string([]byte)) causes the string to escape to
// the heap.
func atoUint32(bytes []byte) (uint32, bool) {
	if len(bytes) == 0 {
		return 0, false
	}
	val := uint32(0)
	for _, b := range bytes {
		if val >= cutoff {
			if val > cutoff {
				return 0, false
			}
			if b > rest {
				return 0, false
			}
		}
		if b < byte('0') || b > byte('9') {
			return 0, false
		}
		val = val*10 + uint32(b-byte('0'))
	}
	return val, true
}

// A Query provides the means to read aggregated transactions from the
// log.  A Query must be created from a Cursor using the NewQuery
// function.
type Query struct {
	cursor     *Cursor
	incomplete map[uint32]Tx
	grp        Grouping
}

var str2type = map[string]TxType{
	"sess":  Sess,
	"req":   Req,
	"bereq": BeReq,
}

var str2reason = map[string]Reason{
	"HTTP/1":  HTTP1,
	"rxreq":   RxReq,
	"esi":     ESI,
	"restart": Restart,
	"pass":    Pass,
	"fetch":   Fetch,
	"bgfetch": BgFetch,
	"pipe":    Pipe,
}

func (c *Cursor) rawTxGrp() []Tx {
	rec := c.Record()
	return []Tx{Tx{
		Type:       TxRaw,
		Reason:     ReasonUnknown,
		Level:      0,
		VXID:       uint32(rec.VXID),
		ParentVXID: 0,
		Records:    []Record{rec},
	}}
}

// NewQuery creates a Query from a Cursor. The Grouping argument indicates
// how transactions from the log should be aggregated: VXID, Request, Session
// or Raw grouping.
//
// XXX: request and session grouping not yet implement, nor are VSL
// queries
func (c *Cursor) NewQuery(grp Grouping, query string) (*Query, error) {
	if c == nil || c.cursor == nil {
		return nil, errors.New("Cursor is nil or uninitialized")
	}
	if err := c.log.checkNil(); err != nil {
		return nil, err
	}
	txMap := make(map[uint32]Tx)
	return &Query{cursor: c, grp: grp, incomplete: txMap}, nil
}

// NextTxGroup returns the next group of transactions from the
// log. Transactions are aggregated according to the Grouping argument
// given in NewQuery().
//
// The Status return value indicates whether there was an error
// reading the log, whether there was an end condition (EOL or EOF),
// or if there are more transactions to be read from the log.
func (q *Query) NextTxGroup() ([]Tx, Status) {
	for {
		status := q.cursor.Next()
		if status != More {
			return nil, status
		}
		if q.grp == GRaw {
			return q.cursor.rawTxGrp(), status
		}

		backend := q.cursor.Backend()
		client := q.cursor.Client()
		if !backend && !client {
			continue
		}

		rec := q.cursor.Record()
		vxid32 := uint32(rec.VXID)
		incmplTx, exists := q.incomplete[vxid32]
		if exists {
			incmplTx.Records = append(incmplTx.Records, rec)
			if rec.Tag == Tag(C.SLT_End) {
				delete(q.incomplete, vxid32)
				txGrp := []Tx{incmplTx}
				return txGrp, status
			}
			q.incomplete[vxid32] = incmplTx
			continue
		}

		tx := Tx{
			VXID:    vxid32,
			Type:    TxUnknown,
			Reason:  ReasonUnknown,
			Level:   1,
			Records: []Record{rec},
		}
		if rec.Tag == Tag(C.SLT_Begin) {
			begin := bytes.Split(rec.Payload, spaceByte)
			typeFromRec := string(begin[0])
			if txtype, exists := str2type[typeFromRec]; exists {
				tx.Type = txtype
			}
			if pvxid, ok := atoUint32(begin[1]); ok {
				tx.ParentVXID = pvxid
			}
			recReason := string(begin[2])
			if reason, exists := str2reason[recReason]; exists {
				tx.Reason = reason
			}
		}
		q.incomplete[vxid32] = tx
	}
}
